import random
import re
import string

INDENT = ' ' * 3

RE_TOKENIZE_STRING = r'(\')(.*?)(\')'
RE_TOKENIZE_COMMENT = r'^([ \t]*?)(#)(.*?)(?=\n|$)'
RE_FUNC = r'(?<=[^\w\d])({})(?=[^\w\d])'
RE_FUNC_PAREN = r'(?<=[^\w\d])({})(\s*)([^\w\d\(])'
RE_OPERATOR = r'(\s*)(\@\<\=|\@\>\=|\@\<\>|\@\=|\@\<|\@\>|\<\=|\>\=|\<\>|\+|\-|\/|\\|\*|\<|\>|\~|\||=|\&|\%)(\s*)'
RE_PARAM_PAREN = r'(\s*)(\(|\))(\s*)'
RE_PARAM_COMMA = r'(\s*)(,)(\s*)'
RE_END = r'(\s*)(;)(\s*?)(\n)'
RE_DIGITS = r'(=|\*|\+|\\|\/|-)(\s*)(\+|\-)(\s*)(\d+)'
RE_INDENT = r'(^)(\s*)(ELSEIF|ENDIF|WHILE|IF|END|ELSE)(.*;$)'
RE_EXECUTEPROCESS = r'(\s*)(EXECUTEPROCESS\(|RUNPROCESS\()(.*)(\);)'

TI_CONTROL_PARAM = (
    'if', 'elseif', 'while', 'end', 'endif', 'break', 'next', 'else', 'end'
)

TI_FUNCTIONS = (
    'attrn', 'attrs', 'cubeattrn', 'cubeattrs', 'dimensionattrn', 'dimensionattrs', 'elementattrn', 'elementattrs',
    'consolidatedavg', 'consolidatedcount', 'consolidatedcountunique', 'consolidatedmax', 'consolidatedmin',
    'isundefinedcellvalue', 'undef', 'undefinedcellvalue', 'date', 'dates', 'day', 'dayno', 'month', 'now', 'time',
    'timst', 'timvl', 'today', 'year', 'dimix', 'dimnm', 'dimsiz', 'dnext', 'dnlev', 'dtype', 'tabdim', 'elcomp',
    'elcompn', 'elementcomponent', 'elementcomponentcount', 'elementcount', 'elementfirst', 'elementindex',
    'elementisancestor', 'elementiscomponent', 'elementisparent', 'elementlevel', 'elementname', 'elementnext',
    'elementparent', 'elementparentcount', 'elementtype', 'elementweight', 'elisanc', 'eliscomp', 'elispar', 'ellev',
    'elpar', 'elparn', 'elweight', 'levelcount', 'fv', 'paymt', 'pv', 'continue', 'abs', 'acos', 'asin', 'atan', 'cos',
    'exp', 'int', 'isund', 'ln', 'log', 'max', 'min', 'mod', 'rand', 'round', 'roundp', 'sign', 'sin', 'sqrt', 'tan',
    'capit', 'char', 'code', 'codew', 'delet', 'fill', 'insrt', 'long', 'lower', 'numbr', 'scan', 'str', 'subst',
    'trim', 'upper', 'asciidelete', 'asciioutput', 'numbertostring', 'numbertostringex', 'setinputcharacterset',
    'setoutputcharacterset', 'setoutputescapedoublequote', 'stringtonumber', 'stringtonumberex', 'textoutput', 'attrnl',
    'attrsl', 'attrdelete', 'attrinsert', 'attrputn', 'attrputs', 'choreattrdelete', 'choreattrinsert', 'choreattrn',
    'choreattrnl', 'choreattrputn', 'choreattrputs', 'choreattrs', 'choreattrsl', 'cubeattrdelete', 'cubeattrinsert',
    'cubeattrputn', 'cubeattrputs', 'cubeattrnl', 'cubeattrsl', 'dimensionattrdelete', 'dimensionattrinsert',
    'dimensionattrputn', 'dimensionattrputs', 'dimensionattrnl', 'dimensionattrsl', 'elementattrnl', 'elementattrsl',
    'elementattrputn', 'elementattrputs', 'elementattrinsert', 'elementattrdelete', 'hierarchyattrputn',
    'hierarchyattrputs', 'hierarchyattrn', 'hierarchyattrs', 'hierarchyattrnl', 'hierarchyattrsl',
    'hierarchysubsetattrs', 'hierarchysubsetattrn', 'hierarchysubsetattrsl', 'hierarchysubsetattrnl',
    'hierarchysubsetattrputs', 'hierarchysubsetattrputn', 'hierarchysubsetattrinsert', 'hierarchysubsetattrdelete',
    'processattrdelete', 'processattrinsert', 'processattrn', 'processattrnl', 'processattrputn', 'processattrputs',
    'processattrs', 'processattrsl', 'subsetattrs', 'subsetattrn', 'subsetattrsl', 'subsetattrnl', 'subsetattrputs',
    'subsetattrputn', 'subsetattrinsert', 'subsetattrdelete', 'viewattrdelete', 'viewattrinsert', 'viewattrn',
    'viewattrnl', 'viewattrputn', 'viewattrputs', 'viewattrs', 'viewattrsl', 'choreerror', 'chorequit', 'chorerollback',
    'setchoreverbosemessages', 'addcubedependency', 'cellgetn', 'cellgets', 'cellincrementn', 'cellisupdateable',
    'cellputn', 'cellputproportionalspread', 'cellputs', 'cubecleardata', 'cubecreate', 'cubedestroy',
    'cubedimensioncountget', 'cubeexists', 'cubegetlogchanges', 'cubesavedata', 'cubesetconnparams',
    'cubesetlogchanges', 'cubetimelastupdated', 'cubeunload', 'cubedatareservationacquire',
    'cubedatareservationrelease', 'cubedatareservationreleaseall', 'cubedatareservationget',
    'cubedatareservationgetconflicts', 'cubedracquire', 'cubedrrelease', 'cubedrreleaseall', 'cubedrget',
    'cubedrgetconflicts', 'formatdate', 'newdateformatter', 'parsedate', 'dimensioncreate',
    'dimensiondeleteallelements', 'dimensiondeleteelements', 'dimensiondestroy', 'dimensionelementcomponentadd',
    'dimensionelementcomponentadddirect', 'dimensionelementcomponentdelete', 'dimensionelementcomponentdeletedirect',
    'dimensionelementdelete', 'dimensionelementdeletedirect', 'dimensionelementexists', 'dimensionelementinsert',
    'dimensionelementinsertdirect', 'dimensionelementprincipalname', 'dimensionexists', 'dimensionhierarchycreate',
    'dimensionsortorder', 'dimensiontimelastupdated', 'dimensiontopelementinsert', 'dimensiontopelementinsertdirect',
    'dimensionupdatedirect', 'createhierarchybyattribute', 'hierarchycontainsallleaves', 'hierarchycreate',
    'hierarchydeleteallelements', 'hierarchydeleteelements', 'hierarchydestroy', 'hierarchyelementcomponentadd',
    'hierarchyelementcomponentadddirect', 'hierarchyelementcomponentdelete', 'hierarchyelementcomponentdeletedirect',
    'hierarchyelementdelete', 'hierarchyelementdeletedirect', 'hierarchyelementexists', 'hierarchyelementinsert',
    'hierarchyelementinsertdirect', 'hierarchyelementprincipalname', 'hierarchyexists', 'hierarchyhasorphanedleaves',
    'hierarchysortorder', 'hierarchytimelastupdated', 'hierarchytopelementinsert', 'hierarchytopelementinsertdirect',
    'hierarchyupdatedirect', 'odbcclose', 'odbcopen', 'odbcopenex', 'odbcoutput', 'setodbcunicodeinterface',
    'executecommand', 'executeprocess', 'getprocesserrorfiledirectory', 'getprocesserrorfilename', 'getprocessname',
    'itemreject', 'itemskip', 'processbreak', 'processerror', 'processexists', 'processexitbychorerollback',
    'processexitbyprocessrollback', 'processquit', 'processrollback', 'runprocess', 'synchronized',
    'cubeprocessfeeders', 'cuberuleappend', 'cuberuledestroy', 'deleteallpersistentfeeders', 'forceskipcheck',
    'ruleloadfromfile', 'getuseactivesandboxproperty', 'serveractivesandboxget', 'serveractivesandboxset',
    'serversandboxclone', 'serversandboxcreate', 'serversandboxesdelete', 'serversandboxdiscardallchanges',
    'serversandboxmerge', 'serversandboxexists', 'serversandboxget', 'serversandboxlistcountget',
    'setuseactivesandboxproperty', 'addclient', 'addgroup', 'assignclienttogroup', 'assignclientpassword',
    'associatecamidtogroup', 'cellsecuritycubecreate', 'cellsecuritycubedestroy', 'deleteclient', 'deletegroup',
    'elementsecurityget', 'elementsecurityput', 'hierarchyelementsecurityget', 'hierarchyelementsecurityput',
    'removecamidassociation', 'removecamidassociationfromgroup', 'removeclientfromgroup', 'sethierarchygroupssecurity',
    'sethierarchyelementgroupssecurity', 'setdimensiongroupssecurity', 'setelementgroupssecurity',
    'securityoverlaygloballockcell', 'securityoverlaycreateglobaldefault', 'securityoverlaydestroyglobaldefault',
    'securityoverlaygloballocknode', 'securityrefresh', 'batchupdatefinish', 'batchupdatefinishwait',
    'batchupdatestart', 'disablebulkloadmode', 'enablebulkloadmode', 'refreshmdxhierarchy', 'savedataall',
    'servershutdown', 'hierarchysubsetaliasget', 'hierarchysubsetaliasset', 'hierarchysubsetcreate',
    'hierarchysubsetdeleteallelements', 'hierarchysubsetdestroy', 'hierarchysubsetelementexists',
    'hierarchysubsetelementdelete', 'hierarchysubsetelementgetindex', 'hierarchysubsetelementinsert',
    'hierarchysubsetexists', 'hierarchysubsetgetsize', 'hierarchysubsetgetelementname', 'hierarchysubsetisallset',
    'hierarchysubsetmdxget', 'hierarchysubsetmdxset', 'publishsubset', 'subsetaliasget', 'subsetaliasset',
    'subsetcreate', 'subsetcreatebymdx', 'subsetdeleteallelements', 'subsetdestroy', 'subsetelementdelete',
    'subsetelementexists', 'subsetelementgetindex', 'subsetelementinsert', 'subsetexists', 'subsetexpandaboveset',
    'subsetformatstyleset', 'subsetgetelementname', 'subsetgetsize', 'subsetisallset', 'subsetmdxget', 'subsetmdxset',
    'publishview', 'disablemtqviewconstruct', 'enablemtqviewconstruct', 'viewcolumndimensionset',
    'viewcolumnsuppresszeroesset', 'viewconstruct', 'viewcreate', 'viewcreatebymdx', 'viewdestroy', 'viewexists',
    'viewextractfilterbytitlesset', 'viewextractskipcalcsset', 'viewextractskipconsolidatedstringsset',
    'viewextractskiprulevaluesset', 'viewextractskipzeroesset', 'viewmdxset', 'viewmdxget', 'viewrowdimensionset',
    'viewrowsuppresszeroesset', 'viewsubsetassign', 'viewsuppresszeroesset', 'viewtitledimensionset',
    'viewtitleelementset', 'viewzeroout', 'addinfocuberestriction', 'executejavan', 'executejavas', 'expand',
    'fileexists', 'logoutput', 'tm1user', 'wildcardfilesearch', 'stringglobalvariable', 'numericglobalvariable'
)

TI_VARIABLES = (
    'DatasourceNameForServer', 'DatasourceNameForClient', 'DatasourceType', 'DatasourceUsername', 'DatasourcePassword',
    'DatasourceQuery', 'DatasourceCubeView', 'DatasourceDimensionSubset', 'DatasourceASCIIDelimiter',
    'DatasourceASCIIDecimalSeparator', 'DatasourceASCIIThousandSeparator', 'DatasourceASCIIQuoteCharacter',
    'DatasourceASCIIHeaderRecords', 'Value_Is_String', 'NValue', 'SValue', 'OnMinorErrorDoItemSkip',
    'MinorErrorLogMax', 'DataSourceODBOCatalog', 'DataSourceODBOConnectionString', 'DataSourceODBOCubeName',
    'DataSourceODBOHierarchyName', 'DataSourceODBOLocation', 'DataSourceODBOProvider', 'DataSourceODBOSAPClientID',
    'DataSourceODBOSAPClientLanguage', 'DataMinorErrorCount', 'MetadataMinorErrorCount', 'ProcessReturnCode',
    'PrologMinorErrorCount'
)


def format_procedure(text):
    token_dict = {}

    text = _tokenize(text, token_dict)

    text = text.replace('\t', '   ')
    text = text.replace(' ', '')
    text = text.replace('\r\n', '\n')

    text = _update_functions(text)
    text = _update_control(text)
    text = _update_operator(text)
    text = _update_variables(text)
    text = _update_indent(text)
    text = _update_executeprocess(text)

    text = _detokenize(text, token_dict)

    return text


def _detokenize(text, token_dict):
    # Replace tokens with their original values
    for k, v in token_dict.items():
        text = text.replace(v, k)

    return text


def _tokenize(text, token_dict):
    # Replace comments with tokens
    found_string = re.search(RE_TOKENIZE_COMMENT, text, flags=re.MULTILINE)
    while found_string:

        start = found_string.start(0)
        end = found_string.end(0)

        entry = ''.join(found_string.groups()).lstrip()

        if entry not in token_dict:
            key = '^COMMENT:' + ''.join(random.choice(string.ascii_uppercase + string.digits) for _ in range(20)) + '^'
            token_dict[entry] = key
        else:
            key = token_dict[entry]

        text = text[:start] + key + text[end:]

        found_string = re.search(RE_TOKENIZE_COMMENT, text, flags=re.MULTILINE)

    # Replace strings with tokens
    found_string = re.search(RE_TOKENIZE_STRING, text, flags=re.DOTALL)
    while found_string:
        start = found_string.start(0)
        end = found_string.end(0)

        entry = ''.join(found_string.groups())

        if entry not in token_dict:
            key = '^STRING:' + ''.join(random.choice(string.ascii_uppercase + string.digits) for _ in range(20)) + '^'
            token_dict[entry] = key
        else:
            key = token_dict[entry]

        text = text[:start] + key + text[end:]

        found_string = re.search(RE_TOKENIZE_STRING, text, flags=re.DOTALL)

    return text


def _update_functions(text):
    text = re.sub(RE_FUNC.format('|'.join(TI_FUNCTIONS)), lambda x: x.group(1).upper(), text, flags=re.IGNORECASE)
    text = re.sub(RE_PARAM_PAREN, r'\2', text)
    text = re.sub(RE_PARAM_COMMA, r'\2 ', text)
    text = re.sub(RE_END, r'\2\4', text)

    return text


def _update_control(text):
    text = re.sub(RE_FUNC.format('|'.join(TI_CONTROL_PARAM)), lambda x: x.group(1).upper(), text, flags=re.IGNORECASE)

    return text


def _update_operator(text):
    text = re.sub(RE_OPERATOR, r' \2 ', text)

    return text


def _update_variables(text):
    for var in TI_VARIABLES:
        text = re.sub(RE_FUNC.format(var), var, text, flags=re.IGNORECASE)

    text = re.sub(RE_DIGITS, r'\1 \3\5', text)

    return text


def _update_indent(text):
    lines_new = []
    indent = 0
    for line in text.split('\n'):
        match = re.match(RE_INDENT, line, flags=re.IGNORECASE)
        if match:
            if match.group(3).upper() in ('IF', 'WHILE'):
                line = INDENT * indent + line
                indent = indent + 1
            elif match.group(3).upper() in ('ELSEIF', 'ELSE'):
                line = INDENT * (indent - 1) + line
            else:
                indent = indent - 1
                line = INDENT * indent + line
        else:
            line = INDENT * indent + line

        line = line.rstrip()

        lines_new.append(line)

    text = '\n'.join(lines_new)

    return text


def _update_executeprocess(text):
    lines_new = []
    for line in text.split('\n'):
        match = re.match(RE_EXECUTEPROCESS, line, flags=re.IGNORECASE)
        if match:
            # Tokenize parameters
            param_string = match.group(3)
            params = []
            level = 0
            start = 0
            for i, char in enumerate(param_string):
                if char == '(':
                    level += 1
                elif char == ')':
                    level -= 1
                elif char == ',' and level == 0:
                    params.append(param_string[start:i])
                    start = i + 1

                if i == len(param_string) - 1:
                    params.append(param_string[start:])

            if len(params) > 3:
                lines_new.append(match.group(1) + match.group(2) + params[0])
                for v, w in zip(params[1::2], params[2::2]):
                    lines_new.append(match.group(1) + INDENT + ',' + v + ',' + w)
                lines_new.append(match.group(1) + match.group(4))
                continue

        lines_new.append(line)

    return '\n'.join(lines_new)



import base64
import json
import logging
import os
import re

from enum import IntEnum, auto

from TM1py.Objects.Cube import Cube
from TM1py.Objects.Dimension import Dimension
from TM1py.Objects.Rules import Rules
from TM1py.Objects.Process import Process
from TM1py.Objects.Hierarchy import Hierarchy


class Operation:

    def __init__(self, source, operation_type, arguments):
        self.source = source
        self.type = operation_type

        self._function = {
            Ops.DELETE_RULE: self._operation_delete_rule,
            Ops.DELETE_VIEW: self._operation_delete_view,
            Ops.DELETE_VIEW_DATA: self._operation_delete_view_data,
            Ops.DELETE_SUBSET: self._operation_delete_subset,
            Ops.DELETE_CUBE: self._operation_delete_cube,
            Ops.DELETE_HIERARCHY: self._operation_delete_hierarchy,
            Ops.DELETE_DIMENSION: self._operation_delete_dimension,
            Ops.DELETE_PROCESS: self._operation_delete_process,
            Ops.DELETE_FILE: self._operation_delete_file,
            Ops.UPDATE_RULE: self._operation_update_rule,
            Ops.UPDATE_CUBE: self._operation_update_cube,
            Ops.UPDATE_VIEW: self._operation_update_view,
            Ops.UPDATE_VIEW_DATA: self._operation_update_view_data,
            Ops.UPDATE_SUBSET: self._operation_update_subset,
            Ops.UPDATE_HIERARCHY: self._operation_update_hierarchy,
            Ops.UPDATE_DIMENSION: self._operation_update_dimension,
            Ops.UPDATE_PROCESS: self._operation_update_process,
            Ops.UPDATE_FILE: self._operation_update_file,
        }[self.type]

        self._arguments = arguments

    def __str__(self):
        return '{}: {}'.format(self.type.name, ', '.join(self._arguments))

    def __repr__(self):
        return '<Operation ({})>'.format(self.__str__())

    def __eq__(self, other):
        return self.type == other.type and self._arguments == other._arguments

    def __hash__(self):
        return hash(self.type.value + str(self._arguments))

    def do(self, target):
        self.target = target
        self._function(*self._arguments)

    def _operation_delete_process(self, process_name):
        """Deletes a turbo integrator process

        Args:
            process_name (str): Process name
        """
        logger = logging.getLogger(__name__)
        logger.info('Called for {}'.format(process_name))

        session = self.target.session
        try:
            if session.processes.exists(process_name):
                session.processes.delete(process_name)
                logger.info('Deleted process {}'.format(process_name))
        except Exception:
            logger.exception('Encountered error while deleting process {}'.format(process_name))
            raise

    def _operation_delete_rule(self, cube_name):
        """Deletes a rule from a cube

        Args:
            cube_name (str): Name of cube to delete rule from
        """
        logger = logging.getLogger(__name__)
        logger.info('Called for {}'.format(cube_name))

        session = self.target.session
        try:
            if session.cubes.exists(cube_name):
                empty_rule = Rules('')
                cube = session.cubes.get(cube_name)
                cube.rules = empty_rule
                session.cubes.update(cube)
                logger.info('Removed rule from cube {}'.format(cube_name))
            else:
                logger.error('Unable to delete rule because cube {} does not exist'.format(cube_name))
        except Exception:
            logger.exception('Encountered error while deleting cube {}'.format(cube_name))
            raise

    def _operation_delete_cube(self, cube_name):
        """Deletes a cube

        Args:
            cube_name (str): Description
        """
        logger = logging.getLogger(__name__)
        logger.info('Called for {}'.format(cube_name))

        session = self.target.session
        try:
            if session.cubes.exists(cube_name):
                session.cubes.delete(cube_name)
                logger.info('Deleted cube {}'.format(cube_name))
        except Exception:
            logger.exception('Encountered error while deleting cube {}'.format(cube_name))
            raise

    def _operation_delete_view(self, cube_name, view_name):
        """Deletes a cube view

        Args:
            cube_name (str): Description
        """
        logger = logging.getLogger(__name__)
        logger.info('Called for {}/{}'.format(cube_name, view_name))

        session = self.target.session
        try:
            if session.cubes.views.exists(cube_name, view_name, False):
                session.cubes.views.delete(cube_name, view_name, False)
                logger.info('Deleted cube view {}/{}'.format(cube_name, view_name))
        except Exception:
            logger.exception('Encountered error while deleting cube view {}/{}'.format(cube_name, view_name))
            raise

    def _operation_delete_view_data(self, cube_name, view_name):
        pass

    def _operation_delete_subset(self, dimension_name, hierarchy_name, subset_name):
        """Deletes a subset

        Args:
            dimension_name (str): Dimension Name
            hierarchy_name (str): Hierarchy Name
            subset_name (str): Subset Name
        """
        logger = logging.getLogger(__name__)
        logger.info('Called for {}/{}/{}'.format(dimension_name, hierarchy_name, subset_name))

        session = self.target.session
        try:
            if session.dimensions.subsets.exists(subset_name, dimension_name, hierarchy_name, False):
                session.dimensions.subsets.delete(subset_name, dimension_name, hierarchy_name, False)
                logger.info('Deleted subset {}/{}/{}'.format(dimension_name, hierarchy_name, subset_name))
        except Exception:
            logger.exception('Encountered error while deleting subset {}/{}/{}'.format(dimension_name, hierarchy_name, subset_name))
            raise

    def _operation_delete_hierarchy(self, dimension_name, hierarchy_name):
        """Deletes a hierarchy from a dimension

        Args:
            dimension_name (str): Dimension name
            hierarchy_name (str): Hierarchy name
        """
        logger = logging.getLogger(__name__)
        logger.info('Called for {}:{}'.format(dimension_name, hierarchy_name))

        session = self.target.session

        try:
            if session.dimensions.hierarchies.exists(dimension_name, hierarchy_name):
                session.dimensions.hierarchies.delete(dimension_name, hierarchy_name)
                logger.info('Deleted hierarchy {}:{}'.format(dimension_name, hierarchy_name))
            else:
                logger.info('Unable to delete hierarchy {}:{} because it does not exist'.format(dimension_name, hierarchy_name))
        except Exception:
            logger.exception('Encountered error while deleting hierarchy {}:{}'.format(dimension_name, hierarchy_name))
            raise

    def _operation_delete_dimension(self, dimension_name):
        """Deletes a dimension

        Args:
            dimension_name (str): Dimension name
        """
        logger = logging.getLogger(__name__)
        logger.info('Called for {}'.format(dimension_name))

        session = self.target.session

        try:
            if session.dimensions.exists(dimension_name):
                session.dimensions.delete(dimension_name)
                logger.info('Deleted dimension {}'.format(dimension_name))
        except Exception:
            logger.exception('Encountered error while deleting dimension {}'.format(dimension_name))
            raise

    def _operation_delete_file(self, file_name):
        """Deletes a file

        Args:
            file_name (str): File name
        """

        logger = logging.getLogger(__name__)
        logger.info('Called for {}'.format(file_name))

        session = self.target.session
        tm1_rest = session._tm1_rest

        try:
            name = 'tm1cm-{}'.format(base64.b32encode(file_name.encode()).decode())
            request = '/api/v1/Contents(\'Blobs\')/Contents(\'{}\')'.format(name)

            tm1_rest.DELETE(request)

            process = 'TAP.File.Delete File'
            params = {
                'Parameters': [
                    {
                        "Name": "pFile",
                        "Value": file_name
                    },
                    {
                        "Name": "pBaseDirectory",
                        "Value": os.path.split(self.target._data_directory)[0]
                    },
                ],
            }

            session.processes.execute(process, params)

        except Exception:
            logger.exception('Encountered error while deleting file {} {}'.format(file_name, name))
            raise

    def _operation_update_rule(self, cube_name):
        """updates a cube rule

        Args:
            cube_name (str): Cube name
        """
        logger = logging.getLogger(__name__)
        logger.info('Called for {}'.format(cube_name))

        session = self.target.session

        try:
            if session.cubes.exists(cube_name):
                cube = session.cubes.get(cube_name)
                cube.rules = Rules(self.source.rules[cube_name]['Rules'])
                session.cubes.update(cube)
            else:
                logger.error('Unable to update cube rule because cube {} does not exist'.format(cube_name))
        except Exception:
            logger.exception('Encountered error while updating rule in cube {}'.format(cube_name))
            raise

    def _operation_update_cube(self, cube_name):
        """Updates a cube

        Args:
            cube_name (str): Cube name
        """
        logger = logging.getLogger(__name__)
        logger.info('Called for {}'.format(cube_name))

        session = self.target.session

        try:
            cube = self.source.cubes[cube_name]
            dimensions = [x['Name'] for x in cube['Dimensions']]

            if session.cubes.exists(cube_name):
                if session.cubes.get(cube_name).dimensions != dimensions:
                    logger.info('Deleting cube {} for being different'.format(cube_name))
                    session.cubes.delete(cube_name)

            if not session.cubes.exists(cube_name):
                cube.setdefault('Rules', '')
                for dimension in dimensions:
                    if not session.dimensions.exists(dimension):
                        session.dimensions.create(Dimension(dimension))
                session.cubes.create(Cube.from_dict(cube))
        except Exception:
            logger.exception('Encountered error while updating cube {}'.format(cube_name))
            raise

    def _operation_update_view(self, cube_name, view_name):
        """Updates a cube

        Args:
            cube_name (str): Cube name
            view_name (str): View name
        """
        logger = logging.getLogger(__name__)
        logger.info('Called for {}/{}'.format(cube_name, view_name))

        session = self.target.session
        rest = session._tm1_rest
        try:
            view = json.dumps(self.source.get_view(cube_name, view_name), ensure_ascii=False)

            if session.cubes.views.exists(cube_name, view_name, False):
                request = '/api/v1/Cubes(\'{}\')/Views(\'{}\')'.format(cube_name, view_name)
                func = rest.PATCH
            else:
                request = '/api/v1/Cubes(\'{}\')/Views'.format(cube_name)
                func = rest.POST

            func(request, view)
        except Exception:
            logger.exception('Encountered error while updating cube view {}/{}'.format(cube_name, view_name))
            raise

    def _operation_update_view_data(self, cube_name, view_name):
        """Updates a cube

        Args:
            cube_name (str): Cube name
            view_name (str): View name
        """
        logger = logging.getLogger(__name__)
        logger.info('Called for {}/{}'.format(cube_name, view_name))

        session = self.target.session
        rest = session._tm1_rest
        try:
            try:
                request = '/api/v1/Cubes(\'{}\')/tm1.Unlock'.format(cube_name)
                rest.POST(request)
                locked = True
            except Exception:
                locked = False

            view_data = self.source.get_view_data(cube_name, view_name)
            for row in view_data:
                groups = [re.match(r'(\[)(.*?)(\]\.\[)(.*?)(\]\.\[)(.*?)(\])', a) for a in row[:-1]]
                dimensions = [g.group(2) for g in groups]
                elements = [g.group(6) for g in groups]
                value = row[-1]

                try:
                    session.cubes.cells.write_value(value, cube_name, elements, dimensions)
                except Exception:
                    logger.error('Error writing row {}:"{}" to {}'.format(elements, value, cube_name))

            if locked:
                request = '/api/v1/Cubes(\'{}\')/tm1.Lock'.format(cube_name)
                rest.POST(request)
        except Exception:
            logger.exception('Encountered error while updating cube view {}/{}'.format(cube_name, view_name))
            raise

    def _operation_update_subset(self, dimension_name, hierarchy_name, subset_name):
        """Updates a subset

        Args:
            dimension_name (str): Dimension Name
            hierarchy_name (str): Hierarchy Name
            subset_name (str): Subset Name
        """
        logger = logging.getLogger(__name__)
        logger.info('Called for {}/{}/{}'.format(dimension_name, hierarchy_name, subset_name))

        session = self.target.session
        rest = session._tm1_rest
        try:
            subset = json.dumps(self.source.get_subset(dimension_name, hierarchy_name, subset_name), ensure_ascii=False)
            if session.dimensions.subsets.exists(subset_name, dimension_name, hierarchy_name, False):
                request = '/api/v1/Dimensions(\'{}\')/Hierarchies(\'{}\')/Subsets(\'{}\')'.format(dimension_name, hierarchy_name, subset_name)
                rest.PATCH(request, subset)
            else:
                request = '/api/v1/Dimensions(\'{}\')/Hierarchies(\'{}\')/Subsets'.format(dimension_name, hierarchy_name)
                rest.POST(request, subset)
        except Exception:
            logger.exception('Encountered error while updating subset {}/{}/{}'.format(dimension_name, hierarchy_name, subset_name))
            raise

    def _operation_update_hierarchy(self, dimension_name, hierarchy_name):
        """Updates a hierarchy on a dimension

        Args:
            dimension_name (str): Dimension Name
            hierarchy_name (str): Hierarchy Name
        """
        logger = logging.getLogger(__name__)
        logger.info('Called for {}:{}'.format(dimension_name, hierarchy_name))

        session = self.target.session

        try:
            data = self.source.get_hierarchy(dimension_name, hierarchy_name)
            hierarchy = Hierarchy.from_dict(data)

            if not session.dimensions.hierarchies.exists(dimension_name, hierarchy_name):
                session.dimensions.hierarchies.create(Hierarchy(hierarchy_name, dimension_name))

            session.dimensions.hierarchies.update_element_attributes(hierarchy)

            target_hierarchy = self.target.get_hierarchy(dimension_name, hierarchy_name)
            if target_hierarchy['Elements'] or data['Elements']:
                session.dimensions.hierarchies.update(hierarchy)

                attributes = [x['Name'] for x in data['ElementAttributes']]
                elements = [x['Name'] for x in data['Elements']]

                if len(attributes) == 0 or len(elements) == 0:
                    return

                attribute_list = ['[}ElementAttributes_%s].[}ElementAttributes_%s].[%s]' % (dimension_name, dimension_name, attribute) for attribute in attributes]
                element_list = ['[{}].[{}].[{}]'.format(dimension_name, hierarchy_name, element) for element in elements]
                mdx = 'SELECT { %s } ON ROWS, { %s } ON COLUMNS FROM [%s]' % (','.join(element_list), ','.join(attribute_list), '}ElementAttributes_' + dimension_name)

                cellset_id = session.cubes.cells.create_cellset(mdx)
                cellset = session.cubes.cells.execute_mdx(mdx, cell_properties=['Ordinal', 'Value', 'Updateable', 'RuleDerived'])

                updates = []
                for element in data['Elements']:
                    for attribute, value in element.get('Attributes', {}).items():
                        if attribute not in attributes:
                            continue
                        cellset_value = cellset.get(
                            ('[%s].[%s].[%s]' % (dimension_name, hierarchy_name, element['Name']), '[}ElementAttributes_%s].[}ElementAttributes_%s].[%s]' % (dimension_name, dimension_name, attribute)), {'Value': '_____'})
                        if value != cellset_value['Value']:
                            if not cellset_value['Updateable'] & 0x10000000:
                                if attribute == 'Format':
                                    value = 'd:' + value
                                update = {
                                    "Ordinal": cellset_value['Ordinal'],
                                    "Value": value
                                }
                                updates.append(update)
                            else:
                                logger.info('Did not update {} for attribute {} in dimension {} because not updateable'.format(element['Name'], attribute, dimension_name))

                if updates:
                    tm1_rest = session._tm1_rest
                    request = "/api/v1/Cellsets('{}')/Cells".format(cellset_id)

                    tm1_rest.PATCH(request, json.dumps(updates, ensure_ascii=False))
                    session.cubes.cells.delete_cellset(cellset_id)
        except Exception:
            logger.exception('Encountered error while updating dimension {}'.format(dimension_name))
            raise

    def _operation_update_dimension(self, dimension_name):
        """Updates a dimension

        Args:
            dimension_name (str): Dimension name
        """
        logger = logging.getLogger(__name__)
        logger.info('Called for {}'.format(dimension_name))

        session = self.target.session
        try:
            data = self.source.dimensions[dimension_name]
            dimension = Dimension.from_dict(data)

            if not session.dimensions.exists(dimension_name):
                session.dimensions.create(dimension)
        except Exception:
            logger.exception('Encountered error while updating dimension {}'.format(dimension_name))
            raise

    def _operation_update_process(self, process_name):
        """Updates a process

        Args:
            process_name (str): Process name
        """

        logger = logging.getLogger(__name__)
        logger.info('Called for {}'.format(process_name))

        session = self.target.session
        try:
            process = self.source.processes[process_name]
            process = Process.from_dict(process)

            if session.processes.exists(process_name):
                session.processes.update(process)
            else:
                session.processes.create(process)

        except Exception:
            logger.exception('Encountered error while updating process {}'.format(process_name))
            raise

    def _operation_update_file(self, file_name):
        """Updates a file

        Args:
            file_name (str): File name
        """

        logger = logging.getLogger(__name__)
        logger.info('Called for {}'.format(file_name))

        session = self.target.session
        tm1_rest = session._tm1_rest

        try:
            name = 'tm1cm-{}'.format(base64.b32encode(file_name.encode()).decode().lower())
            request = '/api/v1/Contents(\'Blobs\')/Contents'
            data = {
                "@odata.type": "#ibm.tm1.api.v1.Document",
                "Name": name
            }

            try:
                tm1_rest.POST(request, json.dumps(data))
            except Exception:
                pass

            request = '/api/v1/Contents(\'Blobs\')/Contents(\'{}\')/Content'.format(name)
            data = self.source.get_file(file_name)

            try:
                url = tm1_rest._base_url + request
                url = url.replace(' ', '%20').replace('#', '%23')

                tm1_rest._s.patch(url=url, headers=tm1_rest._headers, data=data, verify=tm1_rest._verify)
            except Exception:
                logger.exception('unable to patch')

            process = 'TAP.File.Blob to File'
            params = {
                'Parameters': [
                    {
                        "Name": "pBlob",
                        "Value": name
                    },
                    {
                        "Name": "pFile",
                        "Value": file_name
                    },
                    {
                        "Name": "pBaseDirectory",
                        "Value": os.path.split(self.target._data_directory)[0]
                    },
                ],
            }

            session.processes.execute(process, params)
        except Exception:
            logger.exception('Encountered error while updating file {}'.format(file_name))
            raise


class Ops(IntEnum):

    """This enum defines all the possible operations that can take place as part
    of tm1cm against a TM1 application. It also defines the order in which those
    operations must take place.
    """

    DELETE_PROCESS = auto()
    DELETE_RULE = auto()
    DELETE_VIEW_DATA = auto()
    DELETE_VIEW = auto()
    DELETE_SUBSET = auto()
    DELETE_CUBE = auto()
    DELETE_HIERARCHY = auto()
    DELETE_DIMENSION = auto()
    DELETE_FILE = auto()
    UPDATE_PROCESS = auto()
    UPDATE_DIMENSION = auto()
    UPDATE_HIERARCHY = auto()
    UPDATE_CUBE = auto()
    UPDATE_RULE = auto()
    UPDATE_SUBSET = auto()
    UPDATE_VIEW = auto()
    UPDATE_VIEW_DATA = auto()
    UPDATE_FILE = auto()




import logging
import os
import TM1py

from git import Repo
from shutil import rmtree
from tempfile import mkdtemp

from tm1cm.application import LocalApplication, RemoteApplication
from tm1cm.operation import Operation, Ops

AUTHOR = 'tm1cm <tm1cm@local>'


class Migration:
    def __init__(self, source, target):
        logger = logging.getLogger(__name__)

        self.source = source
        self.target = target

        self.path = mkdtemp(prefix='tm1cm_')
        self.repo = Repo.init(self.path)

        if self.target:
            self.target.to_local(self.path, False)
            self.repo.git.add('*')
            self.repo.git.commit('-m', 'initial', author=AUTHOR)

            for folder in ['data', 'scripts', 'files']:
                self._git_rm_path(folder)
        else:
            self.repo.git.commit('--allow-empty', '-m', 'initial')

        self.source.to_local(self.path, True)
        if isinstance(self.source, RemoteApplication):
            self.source = LocalApplication.from_remote(self.source, self.path)

        # Patch TM1py to not delete attributes
        TM1py.Services.HierarchyService._update_element_attributes = _update_element_attributes

        try:
            self.repo.git.add('*')
        except Exception:
            logger.info('Unable to add new files, probably fine')

        self._compute_operations()

    @property
    def operations(self):
        return sorted(self._operations, key=lambda x: x.type)

    def do_operations(self, operations):
        logger = logging.getLogger(__name__)
        retry = []
        for operation in sorted(operations, key=lambda x: x.type):
            try:
                operation.do(self.target)
            except Exception:
                logger.error('Operation {} Failed, will retry at the end'.format(operation))
                retry.append(operation)

        for operation in retry:
            try:
                operation.do(self.target)
            except Exception:
                logger.error('Retry Operation {} Failed'.format(operation))

    def do_all_operations(self):
        self.do_operations(self.operations)

    def _get_operation_list(self, change_type, object_type, name, name_old=None):
        if change_type == 'R':
            a = self._get_operation_list('A', object_type, name)
            b = self._get_operation_list('D', object_type, name_old)
            return a + b
        else:
            action = 'DELETE' if change_type == 'D' else 'UPDATE'
            operation_type = getattr(Ops, '{}_{}'.format(action, object_type.upper()))
            if object_type in ['hierarchy', 'view', 'view_data', 'subset']:
                operation_args = name.split(os.sep)
            else:
                operation_args = (name,)

            return [Operation(self.source, operation_type, operation_args)]

    def _compute_operations(self):
        head_commit = self.repo.head.commit
        self._operations = []
        for diff_added in head_commit.diff(None):
            change_type = diff_added.change_type[:1]

            folder, sub_folder, *object_a = diff_added.a_path.split(os.sep)
            object_a = os.sep.join(object_a)

            try:
                object_a_name, object_a_ext = object_a.rsplit('.', 1)
            except Exception:
                object_a_name = object_a

            _, _, *object_b = diff_added.b_path.split(os.sep)
            object_b = os.sep.join(object_b)

            try:
                object_b_name, object_b_ext = object_b.rsplit('.', 1)
            except Exception:
                object_b_name = object_b

            if folder == 'data':
                op = self._get_operation_list(change_type, object_b_ext, object_b_name, object_a_name)
            else:
                op = self._get_operation_list(change_type, 'file', diff_added.b_path, diff_added.a_path)

            self._operations.extend(op)

    def _git_rm_path(self, folder):
        try:
            self.repo.git.rm('-r', folder)
            rmtree(os.path.join(self.path, folder), ignore_errors=True)
        except Exception:
            pass


def _update_element_attributes(self, hierarchy):
    # get existing attributes first.
    element_attributes = self.elements.get_element_attributes(dimension_name=hierarchy.dimension_name, hierarchy_name=hierarchy.name)
    element_attribute_names = [ea.name for ea in element_attributes]

    # write ElementAttributes that don't already exist !
    for element_attribute in hierarchy.element_attributes:
        if element_attribute not in element_attribute_names:
            self.elements.create_element_attribute(dimension_name=hierarchy.dimension_name, hierarchy_name=hierarchy.name, element_attribute=element_attribute)



import logging
import os
import shutil

FOLDER_STRUCTURE = [
    'config',
    'config/default',
    'config/test',
    'config/staging',
    'config/prod',
    'data',
    'data/cube',
    'data/dimension',
    'data/hierarchy',
    'data/process',
    'data/rule',
    'data/view',
    'data/view_data',
    'data/subset',
    'scripts',
    'scripts/local',
    'scripts/shared',
    'files',
    'files/local',
    'files/shared',
]


def create_scaffold(path):
    """Entry point for create-app.

    Args:
        path (str): Path of the project

    Raises:
        ValueError: Raised if path already exists
    """
    logger = logging.getLogger(__name__)

    logger.debug('create_scaffold called with path: {}'.format(path))
    if os.path.exists(path):
        raise ValueError('path: {} already exists'.format(path))

    try:
        for folder in FOLDER_STRUCTURE:
            new_path = os.path.join(path, folder)
            logger.debug('creating: {}'.format(new_path))
            os.makedirs(new_path)
    except Exception:
        raise

    try:
        logger.debug(__file__)
        include_path = os.path.join(os.path.split(os.path.abspath(__file__))[0], 'include')
        logger.debug(include_path)
        for f in ['connect.yaml', 'tm1cm.yaml', 'credentials.yaml']:
            copy_from = os.path.join(include_path, f)
            copy_to = os.path.join(path, 'config/default', f)
            logger.debug('copying: {} to {}'.format(copy_from, copy_to))
            shutil.copy(copy_from, copy_to)

        copy_from = os.path.join(include_path, '.gitignore')
        copy_to = os.path.join(path, '.gitignore')
        logger.debug('copying: {} to {}'.format(copy_from, copy_to))
        shutil.copy(copy_from, copy_to)
    except Exception:
        raise

    logger.info('created new scaffold at {}'.format(path))



import logging
import os
from fnmatch import fnmatch

import yaml


def get_config(path, config_name, environment=None):
    logger = logging.getLogger(__name__)
    config_path = os.path.join(path, 'config', 'default', config_name + '.yaml')
    logger.debug('Attempting to open file: {}'.format(config_path))
    try:
        config1 = yaml.safe_load(open(config_path))
    except FileNotFoundError:
        config1 = get_default_config(config_name)
    except Exception:
        raise

    if environment:
        config_path = os.path.join(path, 'config', environment, config_name + '.yaml')
        if os.path.isfile(config_path):
            try:
                config2 = yaml.safe_load(open(config_path))
            except FileNotFoundError:
                config2 = {}
            except Exception:
                raise

            return {**config1, **config2}

    return config1


def get_default_config(config_name):
    include_path = os.path.join(os.path.split(os.path.abspath(__file__))[0], 'include')
    path = os.path.join(include_path, config_name + '.yaml')
    with open(path) as fp:
        return yaml.safe_load(fp)


def filter_list(lst, include, exclude, name_func=None, filter_func=None, extra={}):
    if not name_func:
        def name_func(obj, extra={}):
            return str(obj)

    if not filter_func:
        def filter_func(obj, pattern, extra={}):
            result = fnmatch(name_func(obj, extra), pattern)
            return result

    if not isinstance(include, list):
        include = [include]
    if not isinstance(exclude, list):
        exclude = [exclude]

    if isinstance(lst, dict):
        iterator = lst.items()
    else:
        iterator = lst

    include_list = []
    for obj in iterator:
        obj_name = name_func(obj, extra)
        for pattern in include:
            if filter_func(obj, pattern, extra):
                include_list.append(obj_name)
                break

    exclude_list = []
    for obj in iterator:
        obj_name = name_func(obj, extra)
        for pattern in exclude:
            if filter_func(obj, pattern, extra):
                exclude_list.append(obj_name)
                break

    name_list = set(include_list) - set(exclude_list)

    if isinstance(lst, dict):
        return {key: val for key, val in lst.items() if name_func((key, val), extra) in name_list}
    else:
        return [obj for obj in iterator if name_func(obj, extra) in name_list]


class Dumper(yaml.Dumper):

    def increase_indent(self, flow=False, indentless=False):
        return super(Dumper, self).increase_indent(flow, False)



import base64
import copy
import json
import logging
import os
import shutil
from fnmatch import fnmatch
from glob import iglob

import yaml
from TM1py.Objects.NativeView import NativeView
from TM1py.Objects.Subset import AnonymousSubset
from TM1py.Services import TM1Service

from tm1cm.common import filter_list, Dumper
from tm1cm.ti_format import format_procedure


class Application:

    def __init__(self, config):
        self.config = config

        self._cube_object = None
        self._rule_object = None
        self._dimension_object = None
        self._hierarchy_list = None
        self._process_object = None
        self._file_list = []
        self._view_list = []
        self._subset_list = []

        self.refreshed = False

    @property
    def cubes(self):
        return {x['Name']: x for x in self._cube_object}

    @property
    def rules(self):
        return {x['Name']: x for x in self._rule_object}

    @property
    def dimensions(self):
        return {x['Name']: x for x in self._dimension_object}

    @property
    def hierarchies(self):
        return self._hierarchy_list

    @property
    def processes(self):
        return {x['Name']: x for x in self._process_object}

    @property
    def files(self):
        return self._file_list

    @property
    def views(self):
        return self._view_list

    @property
    def subsets(self):
        return self._subset_list

    def refresh(self, overlay=True):
        self._populate(overlay)
        self._filter()

        self.refreshed = True

        return self

    def to_local(self, path, clear=False):
        logger = logging.getLogger(__name__)

        if not self.refreshed:
            self.refresh()

        for folder in ['data', 'scripts', 'files']:
            file_path = os.path.join(path, folder)
            if clear and os.path.exists(file_path):
                shutil.rmtree(file_path)

        # Cube
        file_path = os.path.join(path, 'data', 'cube')
        if self._cube_object:
            os.makedirs(file_path, exist_ok=True)

        for cube in self._cube_object:
            cube_name = cube['Name']
            logger.debug('Processing cube {}'.format(cube_name))

            file_name = '{}.cube'.format(cube_name)
            with open(os.path.join(file_path, file_name), 'wb') as outfile:
                text = self.dump(cube)
                outfile.write(text)

        # View
        file_path = os.path.join(path, 'data', 'view')
        for cube_name, view_name in self._view_list:
            view_path = os.path.join(file_path, cube_name)
            os.makedirs(view_path, exist_ok=True)

            view = self.get_view(cube_name, view_name)
            logger.debug('Processing cube view {}, {}'.format(cube_name, view_name))

            file_name = '{}.view'.format(view_name)
            with open(os.path.join(view_path, file_name), 'wb') as outfile:
                text = self.dump(view)
                outfile.write(text)

        # View Data
        file_path = os.path.join(path, 'data', 'view_data')
        for cube_name, view_name in self._view_data_list:
            view_path = os.path.join(file_path, cube_name)
            os.makedirs(view_path, exist_ok=True)

            view_data = self.get_view_data(cube_name, view_name)
            logger.debug('Processing cube view data {}, {}'.format(cube_name, view_name))

            file_name = '{}.view_data'.format(view_name)
            with open(os.path.join(view_path, file_name), 'wb') as outfile:
                text = self.dump(view_data)
                outfile.write(text)

        # Subsets
        file_path = os.path.join(path, 'data', 'subset')
        for dimension_name, hierarchy_name, subset_name in self._subset_list:
            subset_path = os.path.join(file_path, dimension_name, hierarchy_name)
            os.makedirs(subset_path, exist_ok=True)

            subset = dict(self.get_subset(dimension_name, hierarchy_name, subset_name))
            logger.debug('Processing subset {}/{}/{}'.format(dimension_name, hierarchy_name, subset_name))

            file_name = '{}.subset'.format(subset_name)
            with open(os.path.join(subset_path, file_name), 'wb') as outfile:
                text = self.dump(subset)
                outfile.write(text)

        # Rule
        file_path = os.path.join(path, 'data', 'rule')
        if self._rule_object:
            os.makedirs(file_path, exist_ok=True)

        for cube in self._rule_object:
            cube_name = cube['Name']
            logger.debug('Processing rule {}'.format(cube_name))

            rule = cube.get('Rules', None)
            if rule:
                file_name = '{}.rule'.format(cube_name)
                with open(os.path.join(file_path, file_name), 'wb') as outfile:
                    outfile.write(rule.encode('utf8'))

        # Dimension
        file_path = os.path.join(path, 'data', 'dimension')
        if self._dimension_object:
            os.makedirs(file_path, exist_ok=True)

        for dimension in self._dimension_object:
            dimension_name = dimension['Name']
            logger.debug('Processing dimension {}'.format(dimension_name))

            file_name = '{}.dimension'.format(dimension_name)
            with open(os.path.join(file_path, file_name), 'wb') as outfile:
                text = self.dump(dimension)
                outfile.write(text)

        # Hierarchy
        file_path = os.path.join(path, 'data', 'hierarchy')
        for dimension_name, hierarchy_name in self._hierarchy_list:
            hierarchy_path = os.path.join(file_path, dimension_name)
            os.makedirs(hierarchy_path, exist_ok=True)

            hierarchy = self.get_hierarchy(dimension_name, hierarchy_name)

            file_name = '{}.hierarchy'.format(hierarchy_name)
            with open(os.path.join(hierarchy_path, file_name), 'wb') as outfile:
                hierarchy['ElementAttributes'] = sorted(hierarchy['ElementAttributes'], key=lambda x: x['Name'])

                text = self.dump(hierarchy)
                outfile.write(text)

        # Process
        file_path = os.path.join(path, 'data', 'process')
        if self._process_object:
            os.makedirs(file_path, exist_ok=True)

        procedures = ['PrologProcedure', 'MetadataProcedure', 'DataProcedure', 'EpilogProcedure']

        for process in copy.deepcopy(self._process_object):
            file_name = '{}.process'.format(process['Name'])

            with open(os.path.join(file_path, file_name), 'wb') as outfile:
                output = []
                for procedure in procedures:
                    header = globals().get('{}_BEGIN'.format(procedure[:-9].upper()))

                    text = process[procedure]
                    text = self._fix_generated_lines(text)
                    text = text.replace('\r\n', '\n')

                    footer = globals().get('{}_END'.format(procedure[:-9].upper()))
                    output.append((header + text + footer).encode('utf8'))

                header = globals().get('PROPERTIES_BEGIN').encode('utf8')
                text = self.dump({key: value for key, value in process.items() if key not in procedures})
                footer = globals().get('PROPERTIES_END').encode('utf8')
                output.insert(0, header + text + footer)

                outfile.write(''.encode('utf8').join(output))

        # Files & Scripts
        for obj in self._file_list:
            file_path, file_name = os.path.split(os.path.join(path, os.sep.join(obj)))

            os.makedirs(file_path, exist_ok=True)

            with open(os.path.join(file_path, file_name), 'wb') as outfile:
                outfile.write(self.get_file(os.sep.join(obj)))

    def to_remote(self, session):
        pass

    def _populate(self, overlay=True):
        """ To be overridden """
        pass

    def get_hierarchy(self, dimension, hierarchy=None):
        """ To be overridden """

    def _filter(self):
        c = self.config

        include = c.get('include_cube', '*')
        exclude = c.get('exclude_cube', '')
        self._cube_object = filter_list(self._cube_object, include, exclude, name_func=lambda x, extra: x['Name'])
        self._rule_object = filter_list(self._rule_object, include, exclude, name_func=lambda x, extra: x['Name'])
        self._view_list = filter_list(self._view_list, include, exclude, name_func=lambda x, extra: x[0])

        include = c.get('include_cube_view', '*/*')
        exclude = c.get('exclude_cube_view', '')
        self._view_list = filter_list(self._view_list, include, exclude, name_func=lambda x, extra: '/'.join(x))

        include = c.get('include_cube_view_data', '')
        exclude = c.get('exclude_cube_view_data', '')
        self._view_data_list = filter_list(self._view_list, include, exclude, name_func=lambda x, extra: '/'.join(x))

        include = c.get('include_rule', '*')
        exclude = c.get('exclude_rule', '')
        self._rule_object = filter_list(self._rule_object, include, exclude, name_func=lambda x, extra: x['Name'])

        include = c.get('include_dimension', '*')
        exclude = c.get('exclude_dimension', '')
        self._dimension_object = filter_list(self._dimension_object, include, exclude, name_func=lambda x, extra: x['Name'])
        self._hierarchy_list = filter_list(self._hierarchy_list, include, exclude, name_func=lambda x, extra: x[0])
        self._subset_list = filter_list(self._subset_list, include, exclude, name_func=lambda x, extra: x[0])

        for dimension in self._dimension_object:
            include = c.get('include_dimension_attribute', '*/*')
            exclude = c.get('exclude_dimension_attribute', '')

            dimension['Attributes'] = filter_list(dimension['Attributes'], include, exclude, name_func=lambda x, extra: '{}/{}'.format(extra, x[0]), extra=dimension['Name'])

        include = c.get('include_dimension_hierarchy', '*/*')
        exclude = c.get('exclude_dimension_hierarchy', '')
        self._hierarchy_list = filter_list(self._hierarchy_list, include, exclude, name_func=lambda x, extra: '/'.join(x))
        self._subset_list = filter_list(self._subset_list, include, exclude, name_func=lambda x, extra: '/'.join(x[:2]))

        include = c.get('include_dimension_hierarchy_subset', '*/*/*')
        exclude = c.get('exclude_dimension_hierarchy_subset', '')
        self._subset_list = filter_list(self._subset_list, include, exclude, name_func=lambda x, extra: '/'.join(x))

        include = c.get('include_process', '*')
        exclude = c.get('exclude_process', '')
        self._process_object = filter_list(self._process_object, include, exclude, name_func=lambda x, extra: x['Name'])

        include = c.get('include_file', '')
        exclude = c.get('exclude_file', '')
        self._file_list = filter_list(self._file_list, include, exclude, name_func=lambda x, extra: '/'.join(x))

    def _filter_hierarchy(self, dimension, hierarchy):
        # Filter elements for certain hierarchies
        include = self.config.get('include_dimension_hierarchy_element', '*/*')
        exclude = self.config.get('exclude_dimension_hierarchy_element', '')

        if not isinstance(include, list):
            include = [include]

        if not isinstance(exclude, list):
            exclude = [exclude]

        name = '{}/{}'.format(dimension, hierarchy['Name'])
        include_element = any([fnmatch(name, x) for x in include]) and not any([fnmatch(name, x) for x in exclude])
        if not include_element:
            hierarchy['Elements'] = []
            hierarchy['Edges'] = []
            hierarchy['DefaultMember'] = {}

        hierarchy.setdefault('Elements', [])
        hierarchy.setdefault('Edges', [])
        hierarchy.setdefault('Subsets', [])
        hierarchy.setdefault('DefaultMember', {})

        # Filter dimension hierarchy attributes
        include = self.config.get('include_dimension_hierarchy_attribute', '*/*/*')
        exclude = self.config.get('exclude_dimension_hierarchy_attribute', '')

        def name_func(obj, extra):
            return '{}/{}/{}'.format(extra['dimension_name'], extra['hierarchy_name'], obj['Name'])

        extra = {
            "dimension_name": dimension,
            "hierarchy_name": hierarchy['Name']
        }

        hierarchy['ElementAttributes'] = filter_list(hierarchy['ElementAttributes'], include, exclude, name_func=name_func, extra=extra)

        # Filter dimension hierarchy attributes from Elements
        def name_func(obj, extra):
            return '{}/{}/{}'.format(extra['dimension_name'], extra['hierarchy_name'], obj[0])

        extra = {
            "dimension_name": dimension,
            "hierarchy_name": hierarchy['Name']
        }

        for i, element in enumerate(hierarchy['Elements']):
            lst = filter_list(element['Attributes'], include, exclude, name_func=name_func, extra=extra)
            hierarchy['Elements'][i]['Attributes'] = lst

        # Filter dimension hierarchy attributes values
        include = self.config.get('include_dimension_hierarchy_attribute_value', '*/*/*')
        exclude = self.config.get('exclude_dimension_hierarchy_attribute_value', '')

        for i, element in enumerate(hierarchy['Elements']):
            lst = filter_list(element['Attributes'], include, exclude, name_func=name_func, extra=extra)
            hierarchy['Elements'][i]['Attributes'] = lst

    def _fix_generated_lines(self, text):
        try:
            split = text.split('\r\n')
            index = split.index('#****End: Generated Statements****')

            if index + 1 == len(split) or split[index + 1] != '':
                split.insert(index + 1, '')

            return '\r\n'.join(split)
        except Exception:
            return text

    def dump(self, data):
        if self.config.get('text_output_format', 'YAML').upper() == 'YAML':
            text = yaml.dump(data, Dumper=Dumper, width=255)
        else:
            text = json.dumps(data, indent=4, sort_keys=True, ensure_ascii=False)

        return text.encode('utf8')

    def load(self, fp):
        if self.config.get('text_output_format', 'YAML').upper() == 'YAML':
            data = yaml.safe_load(fp)
        else:
            data = json.safe_load(fp, indent=4, sort_keys=True, ensure_ascii=False)

        return data


class RemoteApplication(Application):
    def __init__(self, config, session=None, session_config=None):
        if session:
            self._connected = True
            self._address = session._tm1_rest._address
            self._address = session._tm1_rest._port
            self._session = session
            self._update_config()
        else:
            self._connected = False
            self._base_url = session_config.get('base_url', '')
            self._address = session_config.get('address', '')
            self._port = session_config.get('port', '')
            self._session_config = session_config

        super().__init__(config)

    def __str__(self):
        return 'url={} connected={}'.format(self._base_url, self._connected)

    def __repr__(self):
        return '<RemoteApplication ({})>'.format(self.__str__())

    @property
    def session(self):
        self.connect()
        return self._session

    def _populate(self, overlay=None):
        logger = logging.getLogger(__name__)

        self.connect()

        rest = self._session._tm1_rest

        c = self.config

        # Cubes & Rules
        try:
            request = '/api/v1/Cubes?$select=Name,Dimensions,Rules&$expand=Dimensions($select=Name)'
            response = rest.GET(request)
            result = json.loads(response.text)['value']

            self._cube_object = result

            # Remove sandbox dimension if present
            for cube in result:
                if cube['Dimensions'][0]['Name'] == 'Sandboxes':
                    cube['Dimensions'].remove(cube['Dimensions'][0])

            self._rule_object = [x.copy() for x in self._cube_object if x['Rules']]

            for cube in self._cube_object:
                if 'Rules' in cube:
                    cube.pop('Rules')
        except Exception:
            logger.exception('Exception occurred when populating cube and rule objects')
            raise

        # Views
        try:
            request = '/api/v1/Cubes?$select=Name,Views&$expand=Views($select=Name)'
            response = rest.GET(request)
            result = json.loads(response.text)['value']

            self._view_list = []
            for cube in result:
                for view in cube['Views']:
                    self._view_list.append((cube['Name'], view['Name']))
        except Exception:
            logger.exception('Exception occurred when cube views')
            raise

        # Subsets
        try:
            request = '/api/v1/Dimensions?$select=Name,Hierarchies&$expand=Hierarchies($select=Name,Subsets;$expand=Subsets($select=Name))'
            response = rest.GET(request)
            result = json.loads(response.text)['value']

            self._subset_list = []
            for dimension in result:
                for hierarchy in dimension['Hierarchies']:
                    for subset in hierarchy['Subsets']:
                        self._subset_list.append((dimension['Name'], hierarchy['Name'], subset['Name']))
        except Exception:
            logger.exception('Exception occurred when populating subset list')
            raise

        # Dimensions & Hierarchies
        try:
            request = '/api/v1/Dimensions?$select=Name,UniqueName,Attributes,Hierarchies&$expand=Hierarchies($expand=Elements($count;$top=0);$select=Name)'
            response = rest.GET(request)
            result = json.loads(response.text)['value']

            self._dimension_object = result

            self._hierarchy_list = []
            for dimension in self._dimension_object:
                for hierarchy in dimension['Hierarchies']:
                    self._hierarchy_list.append((dimension['Name'], hierarchy['Name']))
                # Remove Hierarchies
                dimension['Hierarchies'] = []
        except Exception:
            logger.exception('Exception occurred when populating dimension and hierarchy objects')
            raise

        # Processes
        procedures = ['PrologProcedure', 'MetadataProcedure', 'DataProcedure', 'EpilogProcedure']
        try:
            processes = self._session.processes.get_all()
            self._process_object = [json.loads(x._construct_body()) for x in processes]

            autoformat = c.get('autoformat_ti_process', True)

            for process in self._process_object:
                for procedure in procedures:
                    if autoformat:
                        try:
                            process[procedure] = format_procedure(process[procedure])
                        except Exception:
                            logger.warning('Auto formatting of TI process {} failed for {}. Using original formatting'.format(process['Name'], procedure))

                if 'DataSource' in process:
                    if 'password' in process['DataSource']:
                        del process['DataSource']['password']
        except Exception:
            logger.exception('Exception occurred when populating process objects')
            raise

        # Files
        if c.get('do_file_operations', False):
            try:
                process = c.get('file_to_blob_update_process', None)
                if process:
                    include = c.get('include_file', '')
                    if isinstance(include, str):
                        include = list(include)

                    exclude = c.get('exclude_file', '')
                    if isinstance(exclude, str):
                        exclude = list(exclude)

                    exclude.append('*.pyc')

                    data = {'Parameters': [
                        {'Name': 'pInclude', 'Value': json.dumps(include)},
                        {'Name': 'pExclude', 'Value': json.dumps(exclude)},
                    ]}

                    self._session.processes.execute(process, data)

                request = '/api/v1/Contents(\'Blobs\')/Contents?$select=Name&$filter=startswith(Name, \'tm1cm-\')'
                response = rest.GET(request)
                result = json.loads(response.text)['value']

                self._file_list = [base64.b32decode(x['Name'][6:], casefold=True).decode().split('/') for x in result]
            except Exception:
                logger.exception('Exception occurred when populating file list')
                raise

    def connect(self, session_config=None):
        if session_config:
            self._session_config = session_config

        if not self._connected:
            self._session = TM1Service(**self._session_config)

            self._connected = True
            self._update_config()

    def _update_config(self):
        request = '/api/v1/Configuration'
        response = self._session._tm1_rest.GET(request)
        result = json.loads(response.text)

        self._data_directory = result['DataBaseDirectory']

    def get_hierarchy(self, dimension, hierarchy):
        logger = logging.getLogger(__name__)

        self.connect()

        include = list(self.config.get('include_dimension_hierarchy_element', '*/*'))
        exclude = list(self.config.get('exclude_dimension_hierarchy_element', ''))

        name_comp = '{}/{}'.format(dimension, hierarchy)

        if any([fnmatch(name_comp, x) for x in include]) and not any([fnmatch(name_comp, x) for x in exclude]):
            request = '/api/v1/Dimensions(\'{}\')/Hierarchies(\'{}\')?$select=DefaultMember,ElementAttributes,Name,UniqueName,Edges,Elements&$expand=DefaultMember,ElementAttributes,Edges($select=ParentName,ComponentName,Weight),Elements($select=Name,UniqueName,Index,Type,Attributes)'.format(
                dimension, hierarchy)
        else:
            request = '/api/v1/Dimensions(\'{}\')/Hierarchies(\'{}\')?$select=DefaultMember,ElementAttributes,Name,UniqueName&$expand=DefaultMember,ElementAttributes'.format(dimension, hierarchy)

        try:
            response = self._session._tm1_rest.GET(request)
            result = json.loads(response.text)
            if '@odata.context' in result:
                del result['@odata.context']

            result.setdefault('Elements', [])
            result.setdefault('Edges', [])
            result.setdefault('Subsets', [])

            for element in result['Elements']:
                element.pop('Index')
                element.pop('UniqueName')

            result.pop('DefaultMember')

            self._filter_hierarchy(dimension, result)

            return result
        except Exception:
            logger.exception('Unable to get dimension hierarchy {}/{}'.format(dimension, hierarchy))
            raise

    def get_file(self, file):
        logger = logging.getLogger(__name__)

        self.connect()

        name = 'tm1cm-{}'.format(base64.b32encode(file.encode()).decode().lower())
        request = '/api/v1/Contents(\'Blobs\')/Contents(\'{}\')/Content'.format(name)

        try:
            response = self._session._tm1_rest.GET(request)

            return response.text.encode()
        except Exception:
            logger.exception('Unable to get file {} ({})'.format(file, name))
            raise

    def get_view(self, cube, view):
        logger = logging.getLogger(__name__)

        self.connect()

        view_service = self._session.cubes.views
        try:
            data = view_service.get_native_view(cube, view, False).body
            return json.loads(data)
        except Exception:
            logger.error('Unable to get cube view {}/{}, because it is invalid. Using default view'.format(cube, view))
            return self._construct_empty_view(cube, view)

    def get_view_data(self, cube, view):
        logger = logging.getLogger(__name__)

        self.connect()

        cell_service = self._session.cubes.cells
        rest = self._session._tm1_rest
        try:
            try:
                request = '/api/v1/Cubes(\'{}\')/tm1.Unlock'.format(cube)
                rest.POST(request)
                locked = True
            except Exception:
                locked = False

            data = cell_service.execute_view(cube, view, False, ['Value', 'Updateable'])
            restructed_data = [[*k, v['Value']] for k, v in data.items() if not v['Updateable'] & 0x10000000]

            if locked:
                request = '/api/v1/Cubes(\'{}\')/tm1.Lock'.format(cube)
                rest.POST(request)

            return restructed_data
        except Exception:
            logger.error('Unable to get cube view data for cube={} and view={}'.format(cube, view))
            raise

    def _construct_empty_view(self, cube, view):
        logger = logging.getLogger(__name__)

        self.connect()

        logger.info('Called for {}/{}'.format(cube, view))
        view = NativeView(cube, view, format_string='0.#########')

        rest = self._session._tm1_rest

        request = '/api/v1/Cubes(\'{}\')/Dimensions?$select=Name&$expand=Hierarchies($select=Name,DefaultMember;$expand=DefaultMember($select=Name))'.format(cube)
        response = rest.GET(request)
        result = json.loads(response.text)['value']

        for dimension in result:
            dim = dimension['Name']
            subset = AnonymousSubset(dim, dim, expression='{TM1SUBSETALL( [%s] )}' % dim)
            view.add_title(dimension['Name'], dimension['Hierarchies'][0]['DefaultMember']['Name'], subset)

        return json.loads(view.body, strict=False)

    def get_subset(self, dimension, hierarchy, subset):
        logger = logging.getLogger(__name__)

        self.connect()

        subset_service = self._session.dimensions.subsets
        try:
            return subset_service.get(subset, dimension, hierarchy, False).body_as_dict
        except Exception:
            logger.exception('Unable to get subset {}/{}/{}'.format(dimension, hierarchy, subset))
            raise


class LocalApplication(Application):

    def __init__(self, config, path):
        self.path = path

        super().__init__(config)

    def _populate(self, overlay=True):
        logger = logging.getLogger(__name__)
        c = self.config

        # Cubes & Rules
        try:
            result = []
            path = os.path.join(self.path, 'data', 'cube', '*.cube')
            for filename in iglob(path):
                with open(filename, 'rb') as fp:
                    result.append(self.load(fp))

            self._cube_object = result

            result = []
            path = os.path.join(self.path, 'data', 'rule', '*.rule')
            for filename in iglob(path):
                with open(filename, 'rb') as fp:
                    _, name = os.path.split(filename)
                    name = os.path.splitext(name)[0]
                    result.append({'Name': name, 'Rules': fp.read().decode('utf8')})

            self._rule_object = result
        except Exception:
            logger.exception('Exception occurred when populating cube and rule objects')
            raise

        # Views
        try:
            result = []
            view_path = os.path.join(self.path, 'data', 'view')
            path = os.path.join(view_path, '**', '*.view')
            for filename in iglob(path, recursive=True):
                if os.path.isfile(filename):
                    result.append(filename[len(view_path) + 1:-5].split(os.sep))

                self._view_list = result
        except Exception:
            logger.exception('Exception occurred when populating view objects')
            raise

        # View Data
        try:
            result = []
            view_path = os.path.join(self.path, 'data', 'view_data')
            path = os.path.join(view_path, '**', '*.view_data')
            for filename in iglob(path, recursive=True):
                if os.path.isfile(filename):
                    result.append(filename[len(view_path) + 1:-10].split(os.sep))

                self._view_data_list = result
        except Exception:
            logger.exception('Exception occurred when populating view data objects')
            raise

        # Subsets
        try:
            result = []
            subset_path = os.path.join(self.path, 'data', 'subset')
            path = os.path.join(subset_path, '**', '*.subset')
            for filename in iglob(path, recursive=True):
                if os.path.isfile(filename):
                    result.append(filename[len(subset_path) + 1:-7].split(os.sep))
            self._subset_list = result
        except Exception:
            logger.exception('Exception occurred when populating subset objects')
            raise

        # Dimensions & Hierarchies
        try:
            result = []
            path = os.path.join(self.path, 'data', 'dimension', '*.dimension')
            for filename in iglob(path):
                with open(filename, 'rb') as fp:
                    result.append(self.load(fp))
            self._dimension_object = result

            result = []
            hierarchy_path = os.path.join(self.path, 'data', 'hierarchy')
            path = os.path.join(hierarchy_path, '**', '*.hierarchy')
            for filename in iglob(path, recursive=True):
                if os.path.isfile(filename):
                    result.append(filename[len(hierarchy_path) + 1:-10].split(os.sep))
            self._hierarchy_list = result
        except Exception:
            logger.exception('Exception occurred when populating dimension and hierarchy objects')
            raise

        # Processes
        try:
            result = []
            path = os.path.join(self.path, 'data', 'process', '*.process')

            overlay_options = []
            if c.get('include_ti_queue', False):
                overlay_options.append('queue')
            if c.get('include_ti_overlay', False):
                overlay_options.append('metrics')

            autoformat = c.get('autoformat_ti_process', True)

            for filename in iglob(path):
                with open(filename, 'rb') as fp:
                    process_text = fp.read().decode('utf8')
                    process_properties = self.load(self._get_process_text(process_text, 'PropertiesProcedure'))

                    for procedure in ['PrologProcedure', 'MetadataProcedure', 'DataProcedure', 'EpilogProcedure']:
                        procedure_text = self._get_process_text(process_text, procedure)

                        if autoformat:
                            try:
                                procedure_text = format_procedure(procedure_text)
                            except Exception:
                                logger.warning('Auto formatting of TI process {} failed for {}. Using original formatting'.format(process_properties['Name'], procedure))

                        process_properties[procedure] = procedure_text

                    if 'DataSource' in process_properties:
                        if 'password' in process_properties['DataSource']:
                            del process_properties['DataSource']['password']

                    result.append(process_properties)

            self._process_object = result
        except Exception:
            logger.exception('Exception occurred when populating process objects')
            raise

        # Files
        try:
            result = []
            for folder in ['files', 'scripts']:
                path = os.path.join(self.path, folder, '**', '*')
                for filename in iglob(path, recursive=True):
                    if os.path.isfile(filename):
                        result.append(filename[len(self.path) + 1:].split(os.sep))

            self._file_list = result
        except Exception:
            logger.exception('Exception occurred when populating dimension and hierarchy objects')
            raise

    def get_hierarchy(self, dimension, hierarchy):
        logger = logging.getLogger(__name__)

        path = os.path.join(self.path, 'data', 'hierarchy', dimension)
        filename = '{}.hierarchy'.format(hierarchy)

        try:
            with open(os.path.join(path, filename), 'rb') as fp:
                result = self.load(fp)

                self._filter_hierarchy(dimension, result)
                return result
        except Exception:
            logger.exception('Unable to get hierarchy')
            raise

    def get_file(self, file):
        logger = logging.getLogger(__name__)

        path = os.path.join(self.path, os.sep.join(file.split('/')))

        try:
            with open(path, 'rb') as fp:
                return fp.read()
        except Exception:
            logger.exception('Unable to get file {}'.format(file))
            raise

    def get_view(self, cube, view):
        logger = logging.getLogger(__name__)

        path = os.path.join(self.path, 'data', 'view', cube, '{}.view'.format(view))

        try:
            with open(path, 'rb') as fp:
                return self.load(fp)
        except Exception:
            logger.exception('Unable to get cube view {}/{}'.format(cube, view))
            return {}

    def get_view_data(self, cube, view):
        logger = logging.getLogger(__name__)

        path = os.path.join(self.path, 'data', 'view_data', cube, '{}.view_data'.format(view))

        try:
            with open(path, 'rb') as fp:
                return self.load(fp)
        except Exception:
            logger.exception('Unable to get cube view data {}/{}'.format(cube, view))
            raise

    def get_subset(self, dimension, hierarchy, subset):
        logger = logging.getLogger(__name__)

        path = os.path.join(self.path, 'data', 'subset', dimension, hierarchy, '{}.subset'.format(subset))

        try:
            with open(path, 'rb') as fp:
                return self.load(fp)
        except Exception:
            logger.exception('Unable to get subset {}/{}/{}'.format(dimension, hierarchy, subset))
            raise

    @staticmethod
    def from_remote(app, path):
        remote = LocalApplication(app.config, path)

        remote._cube_object = app._cube_object
        remote._rule_object = app._rule_object
        remote._dimension_object = app._dimension_object
        remote._hierarchy_list = app._hierarchy_list
        remote._process_object = app._process_object
        remote._view_list = app._view_list
        remote._view_data_list = app._view_list
        remote._file_list = app._file_list
        remote._subset_list = app._subset_list

        return remote

    def _get_process_text(self, text, procedure):
        header = globals().get('{}_BEGIN'.format(procedure[:-9].upper()))
        footer = globals().get('{}_END'.format(procedure[:-9].upper()))
        start = text.index(header) + len(header)
        end = text.index(footer)

        process_text = text[start:end]
        process_text = process_text.replace('\r\n', '\n')
        process_text = '\r\n'.join(process_text.split('\n'))

        return process_text


PROPERTIES_BEGIN = """
# =-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-= #
# BEGIN: PROPERTIES                                                           #
# =-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-= #
""".lstrip()

PROPERTIES_END = """
# =-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-= #
# END: PROPERTIES                                                             #
# =-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-= #
""".rstrip()

PROLOG_BEGIN = """
# =-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-= #
# BEGIN: PROLOG PROCEDURE                                                     #
# =-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-= #

"""

PROLOG_END = """

# =-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-= #
# END: PROLOG PROCEDURE                                                       #
# =-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-= #
""".rstrip()

METADATA_BEGIN = """
# =-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-= #
# BEGIN: METADATA PROCEDURE                                                   #
# =-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-= #

"""

METADATA_END = """

# =-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-= #
# END: METADATA PROCEDURE                                                     #
# =-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-= #
""".rstrip()

DATA_BEGIN = """
# =-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-= #
# BEGIN: DATA PROCEDURE                                                       #
# =-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-= #

"""

DATA_END = """

# =-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-= #
# END: DATA PROCEDURE                                                         #
# =-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-= #
""".rstrip()

EPILOG_BEGIN = """
# =-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-= #
# BEGIN: EPILOG PROCEDURE                                                     #
# =-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-= #

"""

EPILOG_END = """

# =-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-= #
# END: EPILOG PROCEDURE                                                       #
# =-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-= #
""".rstrip()



from setuptools import setup, find_packages
import pathlib

here = pathlib.Path(__file__).parent.resolve()

long_description = (here / "README.md").read_text(encoding="utf-8")

VERSION = "0.9.3"

setup(
    name="tm1cm",
    version=VERSION,
    description="A model migration library for TM1 Planning Analytics",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/ajmyers/tm1cm",
    author="Andrew Myers",
    author_email="me@ajmyers.net",
    classifiers=[
        "Development Status :: 5 - Production/Stable",
        "Intended Audience :: Developers",
        "Topic :: Software Development :: Build Tools",
        'License :: OSI Approved :: GNU General Public License v3 or later (GPLv3+)',
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: 3.9",
        "Programming Language :: Python :: 3.10",
        "Programming Language :: Python :: 3 :: Only",
    ],
    keywords="tm1, planning analytics, ibm, pa, olap, cube",
    package_dir={"": "src"},
    packages=find_packages(where="src"),
    python_requires=">=3.7, <4",
    install_requires=[
        "TM1py>=1.2.1",
        "gitdb2==2.0.5",
        "GitPython==2.1.11",
        "termcolor==1.1.0",
        "ntplib==0.3.4",
        "pyyaml==6.0",
    ],
    entry_points={
        'console_scripts': ['tm1cm=tm1cm.__main__:main'],
    },
    project_urls={
        "Bug Reports": "https://github.com/ajmyers/tm1cm/issues",
        "Source": "https://github.com/ajmyers/tm1cm",
    },
    include_package_data=True
)



import cmd
import logging
import os
import subprocess
import sys
import threading
import time
import traceback
from fnmatch import fnmatch

from termcolor import colored

from tm1cm.application import RemoteApplication, LocalApplication
from tm1cm.common import get_config
from tm1cm.migration import Migration


class Interactive(cmd.Cmd):

    def __init__(self, path=None):
        self.intro = 'Welcome to TM1 Code Migrate interactive!\n'
        self.prompt = '(tm1cm) $ '
        self.migration = None
        self.apps = {}
        self.stage = []

        if path:
            self.config_apps_from_path(path)

        super().__init__()

    def config_apps_from_path(self, path):
        logger = logging.getLogger(__name__)

        if not os.path.exists(path):
            return

        if os.path.exists(os.path.join(path, 'config', 'default')):
            app_name = os.path.basename(path)
            environments = os.listdir(os.path.join(path, 'config'))
            for environment in environments:
                if environment == 'default':
                    self.add_local_path('{}.{}'.format(app_name, 'local'), path)
                elif not environment.startswith('.'):
                    logger.info('Adding app: {} with environment: {}'.format(app_name, environment))
                    self.add_path('{}.{}'.format(app_name, environment), path, environment)
        else:
            apps = os.listdir(path)
            for app in apps:
                app_path = os.path.join(path, app)
                app_config_path = os.path.join(path, app, 'config')
                if os.path.exists(os.path.join(app_config_path, 'default')):
                    environments = os.listdir(app_config_path)
                    for environment in environments:
                        if environment == 'default':
                            self.add_local_path('{}.{}'.format(app, 'local'), app_path)
                        elif not environment.startswith('.'):
                            logger.info('Adding app: {} with environment: {}'.format(app, environment))
                            self.add_path('{}.{}'.format(app, environment), app_path, environment)

    def do_path(self, args):
        """Adds a project or collection or projects at the specified path"""
        logger = logging.getLogger(__name__)
        try:
            self.config_apps_from_path(args)
            self.do_ls('app')
        except Exception:
            print('Error occurred:', '\n', traceback.format_exc())
            logger.exception('Error occurred')

    def add_local_path(self, name, path):
        config = get_config(path, 'tm1cm', None)
        self.apps[name] = LocalApplication(config, path)

    def add_path(self, name, path, environment):
        config = get_config(path, 'tm1cm', environment)
        credentials = get_config(path, 'credentials', environment)
        connect = get_config(path, 'connect', environment)

        session_config = {**connect, **credentials}

        self.apps[name] = RemoteApplication(config, session_config=session_config)

    def do_do(self, arg):
        """Execute the currently staged operations. Optionally, specify 'all'
        to run all operations bypassing the staging process"""
        logger = logging.getLogger(__name__)
        try:
            if arg == 'all':
                ops = self.migration.operations
                message = 'Perform ALL operations? ({}/{})'.format(len(ops), len(ops))
            else:
                ops = self.stage
                message = 'Perform STAGED operations? ({}/{})'.format(len(ops), len(self.migration.operations))

            while True:
                print(message)
                confirm = input('[y/n]? ')
                if confirm in ['y', 'n']:
                    break

            spinner = Spinner()

            if confirm == 'y':
                do_backup = self.migration.target.config.get('backup_on_migrate', False)
                allow_override = self.migration.target.config.get('backup_allow_override', True)
                backup_process = self.migration.target.config.get('backup_process', None)

                if backup_process:
                    if allow_override:
                        while True:
                            print('Backup target prior to migration?')
                            confirm = input('[y/n]? ')
                            if confirm in ['y', 'n']:
                                do_backup = True if confirm == 'y' else False
                                break
                    if do_backup:
                        spinner.start()
                        self.migration.target.session.processes.execute(backup_process)
                        spinner.stop()
                pass

                target = self.migration.target
                for op in ops:
                    print(colored('Doing: {}'.format(str(op)), 'green'))
                    spinner.start()
                    op.do(target)
                    spinner.stop()
        except Exception:
            print('Error occurred:', traceback.format_exc())
            logger.exception('Error occurred')

    def do_ls(self, arg):
        logger = logging.getLogger(__name__)
        try:
            if not arg:
                arg = 'operations'
            if arg in ['app', 'apps', 'applications']:
                for app, _ in self.apps.items():
                    print(app)
            if arg in ['migration', 'operations']:
                if self.migration:
                    for op in self.migration.operations:
                        color = 'green' if op in self.stage else 'red'
                        print(colored(op, color))
            if arg in ['stage']:
                for op in self.stage:
                    print(colored(op, 'green'))
        except Exception:
            print('Error occurred:', '\n', traceback.format_exc())
            logger.exception('Error occurred')

    def do_migrate(self, args):
        logger = logging.getLogger(__name__)
        spinner = Spinner()
        self.stage = []
        try:
            app_from, app_to = args.split(' ', maxsplit=2)

            spinner.start()

            app_from = self.apps[app_from].refresh()
            app_to = self.apps[app_to].refresh()

            self.migration = Migration(app_from, app_to)
        except Exception:
            print('Error occurred:', '\n', traceback.format_exc())
            logger.exception('Error occurred')
        finally:
            spinner.stop()

    def do_add(self, arg):
        logger = logging.getLogger(__name__)
        try:
            if not self.migration:
                print('no migration defined')
                return

            ops = [x for x in self.migration.operations if fnmatch(str(x), arg) and x not in self.stage]

            for op in ops:
                self.stage.append(op)
                print(colored('Add: {}'.format(op), 'green'))

            self.stage = sorted(self.stage, key=lambda x: x.type)
        except Exception:
            print('Error occurred:', '\n', traceback.format_exc())
            logger.exception('Error occurred')

    def do_rm(self, arg):
        logger = logging.getLogger(__name__)
        try:
            if not self.migration:
                print('no migration defined')
                return

            ops = [x for x in self.stage if fnmatch(str(x), arg)]

            for op in ops:
                self.stage.remove(op)
                print(colored('Remove: {}'.format(op), 'red'))

            self.stage = sorted(self.stage, key=lambda x: x.type)
        except Exception:
            print('Error occurred:', '\n', traceback.format_exc())
            logger.exception('Error occurred')

    def do_gui(self, arg):
        logger = logging.getLogger(__name__)
        if not self.migration:
            print('no migration')
            return
        try:
            exe = '/usr/local/bin/github'
            print(self.migration.path)
            subprocess.run([exe, self.migration.path], check=True)
        except Exception:
            print('Error. You may need to install github command line tools from the github desktop application')
            logger.exception('Error running github gui')


class Spinner:
    busy = False
    delay = 0.1

    @staticmethod
    def spinning_cursor():
        while 1:
            for cursor in '|/-\\':
                yield cursor

    def __init__(self, delay=None):
        self.spinner_generator = self.spinning_cursor()
        if delay and float(delay):
            self.delay = delay

    def spinner_task(self):
        while self.busy:
            sys.stdout.write(next(self.spinner_generator))
            sys.stdout.flush()
            time.sleep(self.delay)
            sys.stdout.write('\b')
            sys.stdout.flush()

    def start(self):
        self.busy = True
        threading.Thread(target=self.spinner_task).start()

    def stop(self):
        self.busy = False
        time.sleep(self.delay)



import argparse
import logging
import os
import tempfile

from TM1py.Services import TM1Service

from tm1cm.application import LocalApplication, RemoteApplication
from tm1cm.common import get_config
from tm1cm.interactive import Interactive
from tm1cm.migration import Migration
from tm1cm.scaffold import create_scaffold


def main():
    parser = argparse.ArgumentParser()
    parser.add_argument('--mode', help='', required=False, choices=['get', 'put', 'scaffold', 'interactive'], default='interactive')
    parser.add_argument('--path', help='', required=False, default=os.path.abspath(os.getcwd()))
    parser.add_argument('--log', help='', required=False, default=None)
    parser.add_argument('--environment', help='', required=False, default=None)
    parser.add_argument('--debug', help='', required=False, action='store_true')

    args = parser.parse_args()

    show_splash()

    if args.mode in ['scaffold']:
        setup_logger(args.log, args.path, args.debug)
        create_scaffold(args.path)

    if args.mode in ['get', 'put']:
        setup_logger(args.log, args.path, args.debug)
        function = globals().get(args.mode)

        credentials = get_config(args.path, 'credentials', args.environment)
        connect = get_config(args.path, 'connect', args.environment)

        remote_config = {**connect, **credentials}

        remote_session = TM1Service(**remote_config)

        local_config = get_config(args.path, 'tm1cm', args.environment)
        local_path = args.path

        function(local_config, local_path, remote_session)

    if args.mode in ['interactive']:
        setup_logger(args.log, args.path, args.debug, True)
        interactive(args.path)


def get(tm1cm_config, data_path, remote_session):
    app = RemoteApplication(tm1cm_config, remote_session).refresh(False)
    app.to_local(data_path, clear=True)


def put(config, path, session):
    app_from = LocalApplication(config, path).refresh(True)
    app_to = RemoteApplication(config, session).refresh(None)

    migration = Migration(app_from, app_to)
    migration.do_all_operations()


def interactive(path):
    Interactive(path).cmdloop()


def setup_logger(log, path, debug, nostream=False):
    if not log:
        if not path:
            _, log_path = tempfile.mkstemp(prefix='tm1cm_')
        else:
            basename = os.path.basename(path)
            log_path = os.path.join(os.getcwd(), '{}.log'.format(basename))
    else:
        log_path = log

    if nostream:
        handlers = [logging.FileHandler(os.path.abspath(log_path))]
    else:
        handlers = [
            logging.FileHandler(os.path.abspath(log_path)),
            logging.StreamHandler()
        ]

    # setup logger
    logging.basicConfig(
        level=logging.DEBUG if debug else logging.INFO,
        format='%(asctime)s - %(module)s - %(funcName)s(%(lineno)d) - %(name)s - %(levelname)s - %(message)s',
        handlers=handlers
    )


def show_splash():
    splash = r'''----------------------------------
TM1 Code Migrate
----------------------------------                                                      
 Created By: Andrew Myers (me@ajmyers.net) -- https://www.linkedin.com/in/andrew-myers-3112248/
 Homepage: https://github.com/ajmyers/tm1cm
 Issues & Bugs: https://github.com/ajmyers/tm1cm/issues
-----------------------------------
'''
    print(splash)


if __name__ == '__main__':
    main()
